import numpy as np
import pandas as pd
from Bio import PDB, pairwise2
from Bio.PDB.DSSP import DSSP

from copy import copy

def readPDB(filename, name, chains='A', model=0):

    '''
    Given a PDB code and a directory of where to find the structure, returns
    the distance matrix and residue ids.
    '''

    # Load the structure using Bio.PDB
    structure = PDB.PDBParser(QUIET=True).get_structure(name, filename)

    # The number model to use in the structure; usually there is only one, and this is 0 (default)
    m = structure[model]

    # Which chains to extract - either all chains, or a single chain
    if isinstance(chains, str):
        if chains.lower()=='all':    # Just extract all chains
            structure_list = []
            for ch in m:
                structure_list.append(ProteinStructure(ch, name=name+'-'+ch.id))
            return structure_list

        else:    # If it's a string and not 'All', we assume it's a single chain index
            try:    # A try... except here so that it can tell you if you put in an incorrect chain ID
                ch = m[chains]
            except:
                print(chains, 'is not a valid chain ID in', filename)
                return None

            return ProteinStructure(ch, name=name)

    else:
        # If you pass it a list, it will extract those chains
        structure_list = []
        for ch in m:
            structure_list.append(ProteinStructure(ch, name=name+'-'+ch.id))
        return structure_list


def readAlphaFold(filename, name, pae_file=None):

    structure = Bio.PDB.PDBParser(QUIET=True).get_structure(name, filename)
    prot = ProteinStructure(structure[0]["A"], name=name)

    if type(pae_file)!=None:
        pae_matrix = np.genfromtxt(pae_file)
        prot.pae = pae_matrix

    return prot


letters = {'ALA':'A','ARG':'R','ASN':'N','ASP':'D','CYS':'C','GLU':'E','GLN':'Q','GLY':'G','HIS':'H',
           'ILE':'I','LEU':'L','LYS':'K','MET':'M','PHE':'F','PRO':'P','SER':'S','THR':'T','TRP':'W',
           'TYR':'Y','VAL':'V'}

class ProteinStructure:

    '''
    A class for protein structures, which uses Biopython residue and atom objects but is intended to
    be much easier to use that a Biopython <structure> object, which are just a complete mess.
    '''

    def __init__(self, chain, annotation_type='bfactor', annotate_by='residue', resolution=None, name=None):

        self.chain = chain
        self.residues = [residue for residue in chain]
        self.residue_ids = np.array([residue.id[1] for residue in chain])
        self.sequence = ''.join([letters[residue.resname] for residue in chain if residue.resname in letters])
        self.ordered_sequence = ''.join([letters[residue.resname] for residue in chain if residue.resname in letters and not residue.is_disordered()])

        self.n_residues = len(self.residues)
        self.atoms = np.concatenate([[atom for atom in residue] for residue in self.residues])
        self.n_atoms = len(self.atoms)
        self.xyz = np.array([atom.coord for atom in self.atoms])
        self.dmatrix = None
        self._annotation_type = annotation_type
        self._annotate_by = annotate_by
        self.resolution = resolution
        self.name = name
        if len(self.name) == 4:
            self.pdb = self.name.upper()

        if annotation_type.lower()=='bfactor':

            if 'res' in annotate_by.lower():
                bfactors = []
                for residue in self.residues:
                    bfactors.append(np.mean([atom.bfactor for atom in residue]))
                self.bfactors = np.array(bfactors)
            else:
                self.bfactors = np.array([atom.bfactor for atom in self.atoms])

        elif annotation_type.lower()=='plddm':

            plddms = []
            for residue in self.residues:
                plddms.append(np.max([atom.bfactor for atom in residue]))

            self.plddm = np.array(plddms)

    def __len__(self):
        return self.n_residues

    def __str__(self):
        return "Protein structure "+self.name+" with "+str(self.n_residues)+" residues and "+str(self.n_atoms)+" atoms"

    def __getitem__(self, index):

        if isinstance(index, int):
            try:
                ind = int(np.where(self.residue_ids==index)[0][0])
                return self.residues[ind]
            except:
                return ValueError("No such residue in protein!")

        elif isinstance(index, slice):
            sel = np.array([np.where(self.residue_ids==i)[0][0] for i in index])
            return ProteinStructure(self.residues[sel])

        elif isinstance(index, (list, np.ndarray)):
            return ProteinStructure([self.residues[k] for k in index])

    def __iter__(self):
        return iter(self.residues)

    def select_atoms(self, key):

        if isinstance(key, str):

            # if ' or ' in str:
            #     as = []
            #     for x in key.split(' or '):
            #         as.append([n for n,atom in enumerate(self.atoms) if atom.name==x])
            #     return self.select_atoms(np.concatenate(as))

            atom_sel = [n for n,atom in enumerate(self.atoms) if atom.name==key]
            return self.select_atoms(atom_sel)

        elif isinstance(key, int):
            return self.atoms[key]

        else:
            residues_in_selection = [self.atoms[k].get_parent().id[1] for k in key]
            unique_residues = np.unique(residues_in_selection)
            where_res = [np.where(np.array(residues_in_selection)==res)[0] for res in unique_residues]

            final_residue_list = []

            for n,resn in enumerate(unique_residues):
                res = self.residues[list(self.residue_ids).index(resn)]
                newres = PDB.Residue.Residue(res.id, res.resname, res.segid)

                for atom_index in where_res[n]:
                    newres.add(copy(self.atoms[key[atom_index]]))

                final_residue_list.append(newres)

            return ProteinStructure(final_residue_list, annotation_type=self._annotation_type,
                                    annotate_by=self._annotate_by, resolution=self.resolution, name=self.name)


    def distance_matrix(self, method="CA", recalculate=False):

        if type(self.dmatrix)==type(None) or recalculate==True:

            if method=="CA":
                self.dmatrix= CalcDistanceMatrix(self.select_atoms("CA"))
                self.distance_method = "CA"
            elif method.lower() in ('full', 'all'):
                self.dmatrix = CalcDistanceMatrix(self)
                self.distance_method = "full"

        elif self.distance_method != method:
            if method=="CA":
                self.dmatrix= CalcDistanceMatrix(self.select_atoms("CA"))
                self.distance_method = "CA"
            elif method.lower() in ('full', 'all'):
                self.dmatrix = CalcDistanceMatrix(self)
                self.distance_method = "full"

        return self.dmatrix

    def contacts(self, threshold, dist_threshold=None):

        if type(self.dmatrix)==type(None):
            self.dmatrix = CalcDistanceMatrix(self)

        if type(dist_threshold)==type(None):
            return np.where(self.dmatrix < threshold)
        else:
            s = self.dmatrix.shape[0]
            off_diag = np.abs(np.arange(s)[None,:] - np.arange(s)[:,None]) > dist_threshold
            return np.where((self.dmatrix < threshold)&(off_diag))


    def annotate_secondary_structure(dssp_path='mkdssp', filename='temp.pdb'):

        self.save(self, filename)
        self.ss = PDB.DSSP.DSSP(chain, filename, dssp_path)

    def ordered_residues(self):
        return [res for res in self.residues if not res.is_disordered()]

    def save(self, pdb_file):
        io = PDB.PDBIO()
        io.set_structure(self.chain)
        io.save(pdb_file)

def CalcDistanceMatrix(Structure):

    D = []

    for i in range(len(Structure.xyz)):
        D.append(np.sqrt(np.sum((Structure.xyz[i]-Structure.xyz)**2,axis=1)))

    return np.array(D)



"""
TREEANALYSIS.PY
Sam Berry, last edited March 2022

Functions for analyzing posterior distributions of trees as generated from MrBayes
(may someday have support for other kinds of trees). Primarily uses ete3 to analyze
trees, which has significant functionality above and beyond what Bio.Phylo can do,
but trees can also be coverted to and from these two classes in order to do different
kinds of analysis.

(e.g. I use Bio.Phylo for attaching trees to alignments in MultipleSequenceAlignment.py)
"""

import numpy as np
from Bio import Phylo
from Bio.Phylo import Consensus
from ete3 import Tree
from io import StringIO

def bin_trees(forest, rooted=False):

    """Given a set of ETE3 trees, some of which are identical, returns the:
        -- unique trees from the set
        -- counts of each unique tree
        -- indices in the original list corresponding to each unique trees.

    It does so while avoiding calculating the full matrix of robinson-fould distances
    between every pair for efficiency, under the assumption that this matrix will be full of zeros."""

    # calculate the robinson-fould distances of each tree to the first one
    rfdists = rf_distances(forest[0], forest)

    # many of those may be the same; initialize their indices and add them to a growing list
    tree_bins = [np.where(rfdists==0)[0]]

    # these are the indices of trees that are NOT identical, that we now want to consider.
    inds = np.where(rfdists > 0)[0]

    # we will not pay any more attention to the trees matching architecture #1, and only focus on the others now
    new_forest = [forest[i] for i in inds]

    # iterate until we've identified all of the unique architectures
    while len(new_forest) > 0:

        # do basically the same thing
        rfdists = rf_distances(new_forest[0], new_forest)
        tree_bins.append(inds[np.where(rfdists==0)[0]])
        inds = inds[np.where(rfdists>0)[0]]
        new_forest = [forest[i] for i in inds]

    # pick th first tree of each set as its representative - they'll all be identical
    representative_trees = [forest[tree_bins[k][0]] for k in range(len(tree_bins))]
    counts = np.array([len(tree_bins[k]) for k in range(len(tree_bins))])

    return representative_trees, counts, tree_bins


def read_mrbayes_trprobs(filename):

    file = ''

    with open(filename) as f:
        for line in f:
            file = file+line

    ids = {}
    for i in file.split(';')[1].split('\n'):
        l = i.split(' ')
        if len(l) > 4:
            l = [i for i in l if len(i) > 0]
            ids[l[0]] = l[1].replace(',','')

    trees = []; ps = []; Ps = []

    for line in file.split(';')[2:]:
        x = line[line.find('['):]
        if 'p = ' in x and ',' in x:
            p = float(x.split("&W ")[1].split(']')[0])
            P = float(x.split("P = ")[1].split('] = [')[0])
            tree = x.split(' ')[-1]+';'
            tree = Tree(tree)
            for k,leaf in enumerate(tree.get_leaves()):
                leaf.name = ids[leaf.name]
            trees.append(tree)
            ps.append(p)
            Ps.append(P)

    return Ete3Forest(trees, np.array(ps), np.array(Ps), list(ids.values()))

class Ete3Forest:

    def __init__(self, trees, probs, Probs, names):

        self.trees = trees
        self.p = probs
        self.P = Probs
        self.names = names
        self.N = len(self.trees)
        #self.consensus_trees = None

    def __getitem__(self, index):
        return self.trees[index]

    def __iter__(self):
        return __iter__(self.trees)

    def root_all(self, outgroup):

        for tr in self.trees:
            if 'mid' in outgroup:
                root_point = tr.get_midpoint_outgroup()
                tr.set_outgroup(root_point)
            else:
                tr.set_outgroup(outgroup)

    def marginal(self, leaves, outgroup=None, v=True):

        """Calculate the marginal probabilities of tree topologies considering only a subset of leaves.
           Takes ETE3 trees collected into a Forest object, presumably loaded from a MrBayes .trprobs file.
           Relies on trees representing a converged MCMC sample with burnout phase removed such that values
           represent true posterior marginal probabilities.

           Parameters:
            --leaves: the names of the leaves of interest, as a list. should be >3 (>2 if rooted) or else
                       there is only one meaningful topology. all leaves must be exactly as in the tree.
            optional:
                --outgroup: identify one of your leaves as an outgroup to do rooted RF distances. can also
                            pass "midpoint", in which case each tree will be midpoint rooted.
                --v: verbose, true displays results in ascii, false only returns probs and topologies"""

            # will change later if you passed an outgroup
        rooted=False

        # Prune trees to include only leaves of interest
        pruned_trees = []
        for tree in self.trees:
            tr = tree.copy()  # copy to avoid modifying original, and then
            tr.prune(leaves)   # prune - ete3 is very good at this

            # if you gave an outgroup
            if type(outgroup) != type(None):
                rooted=True
                if outgroup=='midpoint':
                    root_point = tr.get_midpoint_outgroup()
                    tr.set_outgroup(root_point)
                else:
                    tr.set_outgroup(outgroup)
            pruned_trees.append(tr)

        unique_trees, counts, x = bin_trees(pruned_trees, rooted=rooted)

        P = []

        for i in x:

            P.append(np.sum(self.p[np.isin(np.arange(self.N), i)]))

        if v:
            for k,tree in enumerate(unique_trees):
                print("Topology",k,": p = "+str(P[k])[:8], "with ", counts[k], "architectures")
                print(tree.get_ascii())
                print('')
                print('')

        return P, unique_trees

    def to_bio(self):
        return BioForest([tree.write() for tree in self.trees], self.p, self.P, self.names)

class BioForest:

    def __init__(self, trees, probs, Probs, names):

        if isinstance(trees[0], Phylo.Newick.Tree):
            self.trees = trees

        elif isinstance(trees[0], str):
            self.trees = [Phylo.read(StringIO(tree), "newick") for tree in trees]

        else:
            raise TypeError("Reading trees from type", trees[0].type, "not yet implemented")

        self.p = probs
        self.P = Probs
        self.names = names
        self.N = len(self.trees)
        #self.consensus_trees = None

    def __getitem__(self, index):
        return self.trees[index]

    def __iter__(self):
        return __iter__(self.trees)

    def find_consensus(self, method, p_thresh=0.95):

        if "maj" in method.lower():
            return Consensus.majority_consensus(trprobs.trees[:np.where(self.P > p_thresh)[0][0]])

        if "strict" in method.lower():
            return Consensus.strict_consensus(trprobs.trees[:np.where(self.P > p_thresh)[0][0]])

    def root_all(self, clade_names):

        for tree in self.trees:
            ca = tree.common_ancestor(next(tree.find_clades(clade_names[0])), next(tree.find_clades(clade_names[1])))
            tree.root_with_outgroup(ca.clades[1])

def rf_distances(tree_1, tree_list):
    return np.array([tree_1.robinson_foulds(tree_2)[0] for tree_2 in tree_list])
